@@ -1,8 +1,11 @@
+
+import re
 import os
 import xmlrpclib
 import sys
 import util
 import shutil
+import httplib
 import urllib 
 import urllib2
 import time
@@ -11,6 +14,8 @@ import optparse
 import zc.lockfile
 import socket
 import tempfile
+import urlparse
+from pkg_resources import parse_version
 from BeautifulSoup import BeautifulSoup
 from glob import fnmatch
 from md5 import md5
@@ -22,6 +27,8 @@ socket.setdefaulttimeout(timeout)
 
 LOG = None
 
+dev_package_regex = re.compile(r'\ddev[-_]')
+
 class Stats:
     """ This is just for statistics """
     def __init__(self):
@@ -143,13 +150,13 @@ class Package:
                 links.append(href)
         return links
 
-    def _links_external(self, html, filename_matches=None):
+    def _links_external(self, html, filename_matches=None, follow_external_index_pages=False):
         """ pypi has external "download_url"s. We try to get anything
-        from there too. This is really ugly and I'm not sure if there's
-        a sane way.
-        The download_url directs either to a website which contains many
-        download links or directly to a package.
+            from there too. This is really ugly and I'm not sure if there's
+            a sane way.  The download_url directs either to a website which
+            contains many download links or directly to a package.
         """
+
         download_links = set()
         soup = BeautifulSoup(html)
         links = soup.findAll("a")
@@ -157,35 +164,68 @@ class Package:
             if link.renderContents().endswith("download_url"):
                 # we have a download_url!! Yeah.
                 url = link.get("href")
-                if not url: continue
+                if not url:
+                    continue
                 download_links.add(url)
+
         for link in download_links:
             # check if the link points directly to a file
             # and get it if it matches filename_matches
+
             if filename_matches:
                 if self.matches(link, filename_matches):
                     yield link
                     continue
+
                 # fetch what is behind the link and see if it's html.
                 # If it is html, download anything from there.
                 # This is extremely unreliable and therefore commented out.
-                """
-                site = urllib2.urlopen(link)
-                if site.headers.type != "text/html":
-                    continue
 
-                # we have a valid html page now. Parse links and download them.
-                # They have mostly no md5 hash.
-                html = site.read()
-                real_download_links = self._fetch_links(html)
-                for real_download_link in real_download_links:
-                    # build absolute links
-                    real_download_link = urllib.basejoin(site.url, real_download_link)
-                    if not filename_matches or self.matches(real_download_link, filename_matches):
-                        yield(real_download_link)
-                """
+                if follow_external_index_pages:
+                    try:
+                        site = urllib2.urlopen(link)
+                    except Exception, e:
+                        LOG.warn('Unload downloading %s (%s)' % (link, e))
+                        continue
 
-    def _links(self, filename_matches=None, external_links=False):
+                    if site.headers.type != "text/html":
+                        continue
+
+                    # we have a valid html page now. Parse links and download them.
+                    # They have mostly no md5 hash.
+                    html = site.read()
+                    real_download_links = self._fetch_links(html)
+                    candidates = list()
+                    for real_download_link in real_download_links:
+                        # build absolute links
+
+                        real_download_link = urllib.basejoin(site.url, real_download_link)
+                        if not filename_matches or self.matches(real_download_link, filename_matches):
+
+                            # we're not interested in dev packages
+                            if not dev_package_regex.search(real_download_link):
+
+                                # Consider only download links that starts with
+                                # the current package name
+                                filename = urlparse.urlsplit(real_download_link)[2].split('/')[-1]
+                                if not filename.startswith(self.name):
+                                    continue
+
+                                candidates.append(real_download_link)
+
+                    def sort_candidates(url1, url2):
+                        """ Sort all download links by package version """
+                        parts1 = urlparse.urlsplit(url1)[2].split('/')[-1]
+                        parts2 = urlparse.urlsplit(url2)[2].split('/')[-1]
+                        return cmp(parse_version(parts1), parse_version(parts2))
+
+                    # and return the 20 latest files
+                    candidates.sort(sort_candidates)
+                    for c in candidates[-20:][::-1]:
+                        yield c
+
+
+    def _links(self, filename_matches=None, external_links=False, follow_external_index_pages=False):
         """ This is an iterator which returns useful links on files for
             mirroring
         """
@@ -209,7 +249,7 @@ class Package:
             yield (url, hash)
 
         if external_links:
-            for link in self._links_external(remote_index_html, filename_matches):
+            for link in self._links_external(remote_index_html, filename_matches, follow_external_index_pages):
                 yield (link, None)
 
     def matches(self, filename, filename_matches):
@@ -218,8 +258,10 @@ class Package:
                 return True
         return False
 
-    def ls(self, filename_matches=None, external_links=False):
-        links = self._links(filename_matches=filename_matches, external_links=external_links)
+    def ls(self, filename_matches=None, external_links=False, follow_external_index_pages=False):
+        links = self._links(filename_matches=filename_matches, 
+                            external_links=external_links, 
+                            follow_external_index_pages=follow_external_index_pages)
         return [(link[0], os.path.basename(link[0]), link[1]) for link in links]
 
     def _get(self, url, filename, md5_hex=None):
@@ -249,6 +291,22 @@ class Package:
         return self._get(*link)
 
     def content_length(self, link):
+
+        # First try to determine the content-length through
+        # HEAD request in order to save bandwidth
+
+        try:
+            parts = urlparse.urlsplit(link)
+            c = httplib.HTTPConnection(parts[1])
+            c.request('HEAD', parts[2])
+            response = c.getresponse()
+            ct = response.getheader('content-length')
+            if ct is not None:
+                ct = long(ct)
+                return ct
+        except Exception, e:
+            LOG.warn('Could not obtain content-length through a HEAD request from %s (%s)' % (link, e))
+
         try:
             return long(urllib2.urlopen(link).headers.get("content-length"))
         except:
@@ -321,10 +379,21 @@ class Mirror:
         fp.write(footer)
         fp.close()
 
-    def mirror(self, package_list, filename_matches, verbose, cleanup, create_indexes, external_links, base_url):
+    def mirror(self, 
+               package_list, 
+               filename_matches, 
+               verbose, 
+               cleanup, 
+               create_indexes, 
+               external_links, 
+               follow_external_index_pages, 
+               base_url):
         stats = Stats()
         full_list = []
         for package_name in package_list:
+
+            LOG.debug('Processing package %s' % package_name)
+
             try:
                 package = Package(package_name)
             except PackageError, v:
@@ -333,7 +402,7 @@ class Mirror:
                 continue
 
             try:
-                links = package.ls(filename_matches, external_links)
+                links = package.ls(filename_matches, external_links, follow_external_index_pages)
             except PackageError, v:
                 stats.error_404(package_name)
                 LOG.debug("Package not available: %s" % v)
@@ -525,6 +594,7 @@ config_defaults = {
     'verbose': True, # log output
     'log_filename': default_logfile,
     'external_links': False, # experimental external link resolve and download
+    'follow_external_index_pages' : False, # experimental, scan index pages for links
 }
 
 
@@ -569,6 +639,8 @@ def run(args=None):
                       default=False, help='create indexes only (no mirroring)')
     parser.add_option('-e', '--follow-external-links', dest='external_links', action='store_true',
                       default=False, help='Follow and download external links)')
+    parser.add_option('-x', '--follow-external-index-pages', dest='follow_external_index_pages', action='store_true',
+                      default=False, help='Follow external index pages and scan for links')
     options, args = parser.parse_args()
     if len(args) != 1:
         parser.error("No configuration file specified")
@@ -584,6 +656,7 @@ def run(args=None):
     create_indexes = config["create_indexes"] in ("True", "1")
     verbose = config["verbose"] in ("True", "1") or options.verbose
     external_links = config["external_links"] in ("True", "1") or options.external_links
+    follow_external_index_pages = config["follow_external_index_pages"] in ("True", "1") or options.follow_external_index_pages
     log_filename = config['log_filename']
     if options.log_filename:
         log_filename = options.log_filename
@@ -603,5 +676,6 @@ def run(args=None):
                       cleanup, 
                       create_indexes, 
                       external_links, 
+                      follow_external_index_pages,
                       config["base_url"])
 
